package cse425s.hw2.syntax;
import cse425s.hw2.syntax.token.*;
import java.io.*;

public class DefaultLexer implements Lexer
{
    public DefaultLexer(Scanner scanner){
        _scanner = scanner;
        _token   = null;
    }
    
    public Token peek() throws IOException{
        while ( _token == null ){
            if ( _scanner.peek() == '\0' )
                throw new EOFException();
            
            while ( (_scanner.peek()==' ')||(_scanner.peek()=='\n')||(_scanner.peek()=='\t') )
                _scanner.read();
            
            char c=_scanner.peek();
            if ( ((c>='a')&&(c<='z')) || ((c>='A')&&(c<='Z')) || (c=='_') ){
                String file = _scanner.getFileName();
                int line = _scanner.getLineNumber();
                int off  = _scanner.getLineOffset();
                StringBuilder sb = new StringBuilder();
                do{
                    sb.append(_scanner.read());
                    c = _scanner.peek();
                }while( ((c>='a')&&(c<='z')) || ((c>='A')&&(c<='Z')) || (c=='_') || ((c>='0')&&(c<='9')) );
                String lexeme = sb.toString();
                if ( lexeme.equals("define") ){
                    _token = new Define(file,line,off);
                }else if ( lexeme.equals("inherits") ){
                    _token = new Inherits(file,line,off);
                }else if ( lexeme.equals("passwords") ){
                    _token = new Passwords(file,line,off);
                }else if ( lexeme.equals("capabilities") ) {
                    _token = new Capabilities(file,line,off);
                }else if ( lexeme.equals("grant") ){
                    _token = new Grant(file,line,off);
                }else if ( lexeme.equals("to") ){
                    _token = new To(file,line,off);
                }else if ( lexeme.equals("check") ){
                    _token = new Check(file,line,off);
                }else{
                    _token = new Identifier(lexeme,file,line,off);
                }
                
                if ( (file.equals("-") || file.equalsIgnoreCase("stdin")) && (lexeme.equalsIgnoreCase("quit") || lexeme.equalsIgnoreCase("exit")) ){
                    _token = null;
                    throw new EOFException();
                }
                
            }else if ( c == ';' ){
                _token = new Semicolon(_scanner.getFileName(),_scanner.getLineNumber(),_scanner.getLineOffset());
                _scanner.read();
            }else if ( c == ',' ){
                _token = new Comma(_scanner.getFileName(),_scanner.getLineNumber(),_scanner.getLineOffset());
                _scanner.read();
            }else{
                System.err.println("[ERROR] In \""+_scanner.getFileName()+"\" at line "+_scanner.getLineNumber()+" at offset "+_scanner.getLineOffset()+": Unrecognized symbol \'" + _scanner.read() + "\'.");
            }
        }
        return _token;
    }
    
    public Token read() throws IOException{
        Token result=peek();
        _token = null;
        return result;
    }
    
    public boolean has(String lexeme) throws IOException{
        return peek().getLexeme().equals(lexeme);
    }
    
    public boolean has(Class tokentype) throws IOException{
        return tokentype.isInstance(peek());
    }
    
    public Token match(String lexeme) throws IOException{
        Token result=null;
        if ( has(lexeme) ){
            result=read();
        }else{
            System.err.println("[ERROR] In \""+ _token.getFileName()+"\" at line "+_token.getLineNumber()+" at offset "+_token.getLineOffset()+": Expected \""+lexeme+"\", found \""+_token.getLexeme()+"\".");
            throw new java.util.regex.PatternSyntaxException("Expected \""+lexeme+"\", found \""+_token.getLexeme()+"\".",lexeme,_token.getLineNumber());
        }
        return result;
    }
    
    public Token match(Class tokentype) throws IOException{
        Token result=null;
        if ( has(tokentype) ){
            result=read();
        }else{
            System.err.println("[ERROR] In \""+ _token.getFileName()+"\" at line "+_token.getLineNumber()+" at offset "+_token.getLineOffset()+": Expected "+tokentype.getSimpleName().toLowerCase()+", found "+_token.getClass().getSimpleName().toLowerCase()+".");
            throw new java.util.regex.PatternSyntaxException("Expected "+tokentype.getSimpleName().toLowerCase()+", found "+_token.getClass().getSimpleName().toLowerCase()+".",tokentype.toString(),_token.getLineNumber());        
        }
        return result;
    }
    
    private Scanner _scanner;
    private Token   _token;
}
